;;;; regex-extras.scm
;;;; Kon Lovett, Feb '08

(use regex)

(declare
  (usual-integrations)
  (disable-interrupts)
  (generic) ; PCRE options use lotsa bits
  #;(disable-warning var)
  (no-bound-checks)
  (no-procedure-checks-for-usual-bindings)
  (export
    regex-version
    regexp-extra-info-set!
    regexp-extra-info
    regexp-options-set!
    regexp-options
    regexp-info
    regexp-info-nametable
    regex-build-config-info ) )

(cond-expand
 [unsafe
  (eval-when (compile)
    (define-macro (##sys#check-chardef-table . _) '(##core#undefined))
    (define-macro (##sys#check-integer . _) '(##core#undefined))
    (define-macro (##sys#check-blob . _) '(##core#undefined))
    (define-macro (##sys#check-vector . _) '(##core#undefined))
    (define-macro (##sys#check-structure . _) '(##core#undefined))
    (define-macro (##sys#check-range . _) '(##core#undefined))
    (define-macro (##sys#check-pair . _) '(##core#undefined))
    (define-macro (##sys#check-list . _) '(##core#undefined))
    (define-macro (##sys#check-symbol . _) '(##core#undefined))
    (define-macro (##sys#check-string . _) '(##core#undefined))
    (define-macro (##sys#check-char . _) '(##core#undefined))
    (define-macro (##sys#check-exact . _) '(##core#undefined))
    (define-macro (##sys#check-port . _) '(##core#undefined))
    (define-macro (##sys#check-number . _) '(##core#undefined))
    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
 [else] )

;;;

#>#include "pcre.h"<#

;;; Which version of PCRE:

(define regex-version
  (let ([substring substring]
        [substring-index substring-index])
    (lambda ()
      (let ([str ((foreign-lambda nonnull-c-string "pcre_version"))])
        (substring str 0 (substring-index " " str)) ) ) ) )

;;; PCRE Types:

(define-foreign-type pcre (c-pointer "pcre"))
(define-foreign-type nonnull-pcre (nonnull-c-pointer "pcre"))

(define-foreign-type pcre_extra (c-pointer "pcre_extra"))
(define-foreign-type nonnull-pcre_extra (nonnull-c-pointer "pcre_extra"))

(define-foreign-enum (pcre-option unsigned-integer)
  (caseless             PCRE_CASELESS)
  (multiline            PCRE_MULTILINE)
  (dotall               PCRE_DOTALL)
  (extended             PCRE_EXTENDED)
  (anchored             PCRE_ANCHORED)
  (dollar-endonly       PCRE_DOLLAR_ENDONLY)
  (extra                PCRE_EXTRA)
  (notbol               PCRE_NOTBOL)
  (noteol               PCRE_NOTEOL)
  (ungreedy             PCRE_UNGREEDY)
  (notempty             PCRE_NOTEMPTY)
  (utf8                 PCRE_UTF8)
  (no-auto-capture      PCRE_NO_AUTO_CAPTURE)
  (no-utf8-check        PCRE_NO_UTF8_CHECK)
  (auto-callout         PCRE_AUTO_CALLOUT)
  (partial              PCRE_PARTIAL)
  (dfa-shortest         PCRE_DFA_SHORTEST)
  (dfa-restart          PCRE_DFA_RESTART)
  (firstline            PCRE_FIRSTLINE)
  (dupnames             PCRE_DUPNAMES)
  (newline-cr           PCRE_NEWLINE_CR)
  (newline-lf           PCRE_NEWLINE_LF)
  (newline-crlf         PCRE_NEWLINE_CRLF)
  (newline-any          PCRE_NEWLINE_ANY)
  (newline-anycrlf      PCRE_NEWLINE_ANYCRLF)
  (bsr-anycrlf          PCRE_BSR_ANYCRLF)
  (bsr-unicode          PCRE_BSR_UNICODE) )

(define pcre-option-symbols '(
  caseless
  multiline
  dotall
  extended
  anchored
  dollar-endonly
  extra
  notbol
  noteol
  ungreedy
  notempty
  utf8
  no-auto-capture
  no-utf8-check
  auto-callout
  partial
  dfa-shortest
  dfa-restart
  firstline
  dupnames
  newline-cr
  newline-lf
  newline-crlf
  newline-any
  newline-anycrlf
  bsr-anycrlf
  bsr-unicode) )

(define-foreign-enum (pcre-info-field unsigned-int)
  (options          PCRE_INFO_OPTIONS)
  (size             PCRE_INFO_SIZE)
  (capturecount     PCRE_INFO_CAPTURECOUNT)
  (backrefmax       PCRE_INFO_BACKREFMAX)
  (firstbyte        PCRE_INFO_FIRSTBYTE)
  (firstchar        PCRE_INFO_FIRSTCHAR)
  (firsttable       PCRE_INFO_FIRSTTABLE)
  (lastliteral      PCRE_INFO_LASTLITERAL)
  (nameentrysize    PCRE_INFO_NAMEENTRYSIZE)
  (namecount        PCRE_INFO_NAMECOUNT)
  (nametable        PCRE_INFO_NAMETABLE)
  (studysize        PCRE_INFO_STUDYSIZE)
  (default-tables   PCRE_INFO_DEFAULT_TABLES)
  (okpartial        PCRE_INFO_OKPARTIAL)
  (jchanged         PCRE_INFO_JCHANGED)
  (hascrorlf        PCRE_INFO_HASCRORLF) )

(define pcre-info-field-types '(
  (options          . long-integer)
  (size             . long-integer)
  (capturecount     . integer)
  (backrefmax       . integer)
  (firstbyte        . integer)
  (firstchar        . integer)
  (firsttable       . pointer)
  (lastliteral      . integer)
  (nameentrysize    . integer)
  (namecount        . integer)
  (nametable        . pointer)
  (studysize        . integer)
  (default-tables   . pointer)
  (okpartial        . boolean)
  (jchanged         . boolean)
  (hascrorlf        . boolean) ) )

(define pcre-info-field-symbols (map car pcre-info-field-types))

(define-foreign-enum (pcre-config-field unsigned-int)
  (utf8                     PCRE_CONFIG_UTF8)
  (newline                  PCRE_CONFIG_NEWLINE)
  (link-size                PCRE_CONFIG_LINK_SIZE)
  (posix-malloc-threshold   PCRE_CONFIG_POSIX_MALLOC_THRESHOLD)
  (match-limit              PCRE_CONFIG_MATCH_LIMIT)
  (stackrecurse             PCRE_CONFIG_STACKRECURSE)
  (unicode-properties       PCRE_CONFIG_UNICODE_PROPERTIES)
  (match-limit-recursion    PCRE_CONFIG_MATCH_LIMIT_RECURSION)
  (bsr                      PCRE_CONFIG_BSR) )

(define pcre-config-field-types '(
  (utf8                     . boolean)
  (newline                  . integer)
  (link-size                . integer)
  (posix-malloc-threshold   . integer)
  (match-limit              . integer)
  (stackrecurse             . boolean)
  (unicode-properties       . boolean)
  (match-limit-recursion    . integer)
  (bsr                      . boolean) ) )

(define pcre-config-field-symbols (map car pcre-config-field-types))

;;; The regexp structure primitives:

(define re-finalizer
  (foreign-lambda void "pcre_free" c-pointer) )

(define-inline (%regexp? x)
  (##sys#structure? x 'regexp) )

(define-inline (%regexp-code rx)
  (##sys#slot rx 1) )

(define-inline (%regexp-extra rx)
  (##sys#slot rx 2) )

(define-inline (%regexp-options rx)
  (##sys#slot rx 3) )

(define-inline (%regexp-extra-set! rx extra)
  (when extra (set-finalizer! extra re-finalizer))
  (##sys#setslot rx 2 extra) )

(define-inline (%regexp-options-set! rx options)
  (##sys#setslot rx 3 options) )

;;; Optimized compiled regular expression extra structure:

;; Create optimization structure

(define re-extra
  ; Only the flags need to be initialized
  ; When a flag bit is 0 the corresponding field is ignored
  (foreign-lambda* pcre_extra ()
    "pcre_extra *extra = (pcre_extra *) ((pcre_malloc) (sizeof (pcre_extra)));"
    "if (extra) {"
    "  extra->flags = 0;"
    "  extra->study_data = NULL;"
    "  extra->match_limit = 0;"
    "  extra->match_limit_recursion = 0;"
    "  extra->callout_data = NULL;"
    "  extra->tables = NULL;"
    "}"
    "return (extra);"))

;; Set specific field

(define (re-extra-field-set! loc extra key val)
  (case key
    [(match-limit)
      (##sys#check-exact val loc)
      ((foreign-lambda* void ((nonnull-pcre_extra extra) (unsigned-long val))
          "extra->match_limit = val;"
          "extra->flags |= PCRE_EXTRA_MATCH_LIMIT;")
        extra val)]
    [(match-limit-recursion)
      (##sys#check-exact val loc)
      ((foreign-lambda* void ((nonnull-pcre_extra extra) (unsigned-long val))
          "extra->match_limit_recursion = val;"
          "extra->flags |= PCRE_EXTRA_MATCH_LIMIT_RECURSION;")
        extra val)]
    [(callout-data)
      (##sys#check-blob val loc)
      ((foreign-lambda* void ((nonnull-pcre_extra extra) (nonnull-byte-vector val))
          "extra->callout_data = val;"
          "extra->flags |= PCRE_EXTRA_CALLOUT_DATA;")
        extra val)]
    [(tables)
      (##sys#check-chardef-table val loc)
      ((foreign-lambda* void ((nonnull-pcre_extra extra) ((const (c-pointer unsigned-char)) val))
          "extra->tables = val;"
          "extra->flags |= PCRE_EXTRA_TABLES;")
        extra val)]
    [else
      (warning loc "unrecognized extra field key" key)] ) )

;;

(define (re-extra-field loc extra key)
  (case key
    [(match-limit)
      ((foreign-lambda* unsigned-int ((nonnull-pcre_extra extra) (unsigned-long val))
          "return (extra->match_limit);")
        extra)]
    [(match-limit-recursion)
      ((foreign-lambda* unsigned-int ((nonnull-pcre_extra extra) (unsigned-long val))
          "return (extra->match_limit_recursion);")
        extra)]
    [(callout-data)
      ((foreign-lambda* c-pointer ((nonnull-pcre_extra extra) (nonnull-byte-vector val))
          "return (extra->callout_data);")
        extra)]
    [(tables)
      (regex-chardef-table
        ((foreign-lambda* (c-pointer unsigned-char) ((nonnull-pcre_extra extra) ((const (c-pointer unsigned-char)) val))
            "return (extra->tables);")
          extra))]
    [else
      (warning loc "unrecognized extra field key" key)] ) )

;; Set fields

(define (re-extra-fields-set! loc extra . args)
  (let loop ([args args])
    (unless (null? args)
      (let ([key (car args)]
            [val (cadr args)])
        (re-extra-field-set! loc extra key val)
        (loop (cddr args)) ) ) ) )

;; Set user override fields

(define (regexp-extra-info-set! rx . args)
  (##sys#check-structure rx 'regexp 'regexp-extra-info-set!)
  (apply re-extra-fields-set! 'regexp-extra-info-set!
                              (or (%regexp-extra rx)
                                  (let ([extra (re-extra)])
                                    (if extra
                                        (begin
                                          (%regexp-extra-set! rx extra)
                                          extra )
                                        (##sys#error-hook 6 'regexp-extra-info-set!) ) ) )
                              args) )

;; Get user override fields

(define (regexp-extra-info rx . fields)
  (##sys#check-structure rx 'regexp 'regexp-extra-info)
  (let ([extra (%regexp-extra rx)])
    (if extra
        (map
          (lambda (sym)
            (##sys#check-symbol sym 'regexp-extra-info)
            (list sym (re-extra-field 'regexp-extra-info extra sym)) )
          fields)
        '() ) ) )

;;; Regexp options:

;; Get a list of option symbols from the integer options

(define (integer-options->symbols options)
  (let ([lst '()])
    (for-each
      (lambda (sym)
        (unless (zero? (bitwise-and options (pcre-option->number sym)))
          (set! lst (cons sym lst)) ) )
      pcre-option-symbols)
    lst ) )

;; Set the 'exec' options

(define (regexp-options-set! rx . options)
  (##sys#check-structure rx 'regexp 'regexp-options-set!)
  (%regexp-options-set! rx (pcre-option->number options)) )

;; Get the 'exec' options

(define (regexp-options obj)
  (integer-options->symbols
    (cond [(%regexp? obj)   (%regexp-options obj)]
          [(integer? obj)   obj]
          [else
            (##sys#signal-hook #:type-error 'regexp-options
             "bad argument type - not an integer or compiled regular expression" obj)] ) ) )

;;; Regexp 'fullinfo':

(define re-info-long-integer
  (foreign-lambda* long (((const nonnull-pcre) code) (pcre_extra extra) (int fieldno))
    "long int val;"
    "pcre_fullinfo (code, extra, fieldno, &val);"
    "return (val);") )

(define re-info-integer
  (foreign-lambda* int (((const nonnull-pcre) code) (pcre_extra extra) (int fieldno))
    "int val;"
    "pcre_fullinfo (code, extra, fieldno, &val);"
    "return (val);") )

(define re-info-boolean
  (foreign-lambda* bool (((const nonnull-pcre) code) (pcre_extra extra) (int fieldno))
    "int val;"
    "pcre_fullinfo (code, extra, fieldno, &val);"
    "return (val);") )

(define re-info-pointer
  (foreign-lambda* c-pointer (((const nonnull-pcre) code) (pcre_extra extra) (int fieldno))
    "void *val;"
    "pcre_fullinfo (code, extra, fieldno, &val);"
    "return (val);") )

;;

(define (regexp-info rx . fields)
  (##sys#check-structure rx 'regexp 'regexp-info)
  (let ([code (%regexp-code rx)]
        [extra (%regexp-extra rx)])
    (map
      (lambda (sym)
        (##sys#check-symbol sym 'regexp-info)
        (list sym
              (and-let* ([typ (alist-ref sym pcre-info-field-types eq?)])
                (let ([fldno (pcre-info-field->number sym)])
                  (case typ
                    [(boolean)
                      (re-info-boolean code extra fldno)]
                    [(long-integer)
                      (re-info-long-integer code extra fldno)]
                    [(integer)
                      (re-info-integer code extra fldno)]
                    [(pointer)
                      (and-let* ([ptr (re-info-pointer code extra fldno)])
                        (case sym
                          [(default-tables firsttable)
                            (regex-chardef-table ptr)]
                          [else
                            ptr]))]
                    [else
                      (##sys#error 'regexp-info "unknown type" typ) ] ) ) ) ) )
      (if (null? fields) pcre-info-field-symbols fields)) ) )

;;

#>
typedef struct {
  uint16_t cc;
  char name[1];
} pcre_nametable_entry;

typedef struct {
  char *nametable;
  int entrysize;
} pcre_nametable;

#define pcre_nametable_ref(nt, idx) ((pcre_nametable_entry *) ((nt)->nametable + ((idx) * (nt)->entrysize)))

static int
get_nametable_entrycount (const pcre *code, const pcre_extra *extra)
{
  int val;
  pcre_fullinfo (code, extra, PCRE_INFO_NAMECOUNT, &val);
  return (val);
}

static pcre_nametable *
get_nametable (const pcre *code, const pcre_extra *extra)
{
  pcre_nametable *nametable = (pcre_nametable *) ((pcre_malloc) (sizeof (pcre_nametable)));
  if (nametable) {
    pcre_fullinfo (code, extra, PCRE_INFO_NAMETABLE, &nametable->nametable);
    pcre_fullinfo (code, extra, PCRE_INFO_NAMEENTRYSIZE, &nametable->entrysize);
  }
  return (nametable);
}

static char *
get_nametable_entry ( const pcre_nametable *nt, int idx, int *pcc)
{
  pcre_nametable_entry *entry = pcre_nametable_ref (nt, idx);

  /* Number of capturing parentheses is MSB */
# ifdef C_LITTLE_ENDIAN
  uint16_t cc;
  ((uint8_t *) &cc)[0] = ((uint8_t *) &entry->cc)[1];
  ((uint8_t *) &cc)[1] = ((uint8_t *) &entry->cc)[0];
  *pcc = cc;
# else
  *pcc = entry->cc;
# endif

  return (entry->name);
}

#undef pcre_nametable_ref
<#

(define-foreign-type pcre_nametable (c-pointer "pcre_nametable"))

(define-foreign-type nonnull-pcre_nametable (nonnull-c-pointer "pcre_nametable"))

(define re-nametable-entrycount
  (foreign-lambda int "get_nametable_entrycount" (const nonnull-pcre) (const pcre_extra)))

(define re-nametable
  (foreign-lambda pcre_nametable "get_nametable" (const nonnull-pcre) (const pcre_extra)))

(define re-nametable-entry
  (foreign-lambda nonnull-c-string "get_nametable_entry" (const nonnull-pcre_nametable) int (c-pointer int)))

(define (regexp-info-nametable rx)
  (##sys#check-structure rx 'regexp 'regexp-info-nametable)
  (let ([code (%regexp-code rx)]
        [extra (%regexp-extra rx)])
    (let ([cnt (re-nametable-entrycount code extra)])
      (if (fx= 0 cnt)
          '()
          (let ([name-table (re-nametable code extra)])
            (if name-table
                (begin
                  (set-finalizer! name-table re-finalizer)
                  (let loop ([idx 0] [lst '()])
                    (if (fx= cnt idx)
                        lst
                        (let-location ([cc int])
                          (let ([nam (re-nametable-entry name-table idx #$cc)])
                            (loop (fx+ idx 1) (cons (list nam idx cc) lst)) ) ) ) ) )
                (##sys#error-hook 6 'regexp-info-nametable) ) ) ) ) ) )

;;; PCRE config info:

;;

(define config-info-integer
  (foreign-lambda* int ((int fieldno))
    "int val;"
    "pcre_config (fieldno, &val);"
    "return (val);") )

(define config-info-boolean
  (foreign-lambda* bool ((int fieldno))
    "int val;"
    "pcre_config (fieldno, &val);"
    "return (val);") )

;;

(define (regex-build-config-info . fields)
  (map
    (lambda (sym)
      (##sys#check-symbol sym 'regex-build-config-info)
      (list sym
            (and-let* ([typ (alist-ref sym pcre-config-field-types eq?)])
              (let ([fldno (pcre-config-field->number sym)])
                (case typ
                  [(boolean)
                    (config-info-boolean fldno)]
                  [(integer)
                    (let ([int (config-info-integer fldno)])
                      (if (eq? 'newline sym)
                          (integer->char int)
                          int ) ) ]
                  [else
                    (##sys#error 'regex-build-config-info "unknown type" typ) ] ) ) ) ) )
    (if (null? fields) pcre-config-field-symbols fields)) )
